from .base import LogFile
from .constants import (
    LpStatusMemoryLimit,
    LpStatusSolved,
    LpStatusInfeasible,
    LpStatusTimeLimit,
    LpStatusUnbounded,
    LpStatusNotSolved,
)
import re
import numpy as np


class GUROBI(LogFile):
    name = "GUROBI"

    def __init__(self, path, **options):
        super().__init__(path, **options)
        self.solver_status_map = {
            "Optimal solution found": LpStatusSolved,
            "Solved with barrier": LpStatusSolved,
            "Model is infeasible": LpStatusInfeasible,
            "Model is infeasible or unbounded": LpStatusInfeasible,
            "Time limit reached": LpStatusTimeLimit,
            "Out of memory": LpStatusMemoryLimit,
            "ERROR 10001": LpStatusMemoryLimit,
            "ERROR 10003": LpStatusNotSolved,
            "^Model is unbounded": LpStatusUnbounded,
        }
        self.version_regex = r"Gurobi Optimizer version (\S+)"
        self.progress_names = [
            "Node",
            "NodesLeft",
            "Objective",
            "Depth",
            "IInf",
            "BestInteger",
            "CutsBestBound",
            "Gap",
            "ItpNode",
            "Time",
        ]
        self.progress_filter = r"(^[\*H]?\s+\d.*$)"

    def get_cuts(self):
        regex = r"Cutting planes:([\n\s\-\w:]+)Explored"
        result = self.apply_regex(regex, flags=re.MULTILINE)
        if not result:
            # if no cuts found, return empty dictionary
            return {}
        cuts = [r for r in result.split("\n") if r != ""]
        regex = r"\s*{}: {}".format(self.wordSearch, self.numberSearch)
        searches = [re.search(regex, v) for v in cuts]
        return {
            s.group(1): int(s.group(2))
            for s in searches
            if s is not None and s.lastindex >= 2
        }

    def get_matrix(self):
        regex = r"Optimize a model with {0} rows, {0} columns and {0} nonzeros".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="int")

    def get_matrix_post(self):
        regex = r"Presolved: {0} rows, {0} columns, {0} nonzeros".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="int")

    def get_stats(self):
        regex = r"{1}( \(.*\))?\n(Warning:.*\n)?Best objective ({0}|-), best bound ({0}|-), gap ({0}|-)".format(
            self.numberSearch, self.wordSearch
        )
        # content_type = ['', '', 'float', 'float', 'float']
        solution = self.apply_regex(regex)
        if solution is None:
            return None, None, None, None

        status = solution[0]
        objective, bound, gap_rel = [
            float(solution[pos]) if solution[pos] != "-" else None for pos in [3, 5, 7]
        ]
        return status, objective, bound, gap_rel

    def get_cuts_time(self):
        progress = self.get_progress()
        if not len(progress):
            return None
        df_filter = np.all(
            (
                progress.Node.str.match(r"^\*?H?\s*0"),
                progress.NodesLeft.str.match(r"^\+?H?\s*2"),
            ),
            axis=0,
        )

        cell = progress.Time.iloc[0]
        if len(df_filter) and any(df_filter):
            # we finished the cuts phase
            cell = progress.Time[df_filter].iloc[0]

        number = re.search(self.numberSearch, cell).group(1)
        return float(number)

    def get_lp_presolve(self):
        """
        :return: tuple  of length 3
        """
        regex = r"Presolve time: {0}s".format(self.numberSearch)
        time = self.apply_regex(regex, pos=0, content_type="float")
        if time is None:
            time = None
        regex = r"Presolve removed {0} rows and {0} columns".format(self.numberSearch)
        result = self.apply_regex(regex, content_type="int")
        if result is None:
            result = None, None
        return {"time": time, "rows": result[0], "cols": result[1]}

    def get_time(self):
        regex = r"Explored {0} nodes \({0} simplex iterations\) in {0} seconds".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="float", pos=2)

    def get_nodes(self):
        regex = r"Explored {0} nodes \({0} simplex iterations\) in {0} seconds".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="float", pos=0)

    def get_root_time(self):
        regex = r"Root relaxation: objective {0}, {0} iterations, {0} seconds".format(
            self.numberSearch
        )
        return self.apply_regex(regex, pos=2, content_type="float")

    def process_line(self, line):
        keys = [
            "n",
            "n_left",
            "obj",
            "iinf",
            "b_int",
            "b_bound",
            "ItCnt",
            "gap",
            "depth",
            "time",
        ]
        args = {k: self.numberSearch for k in keys}
        args["gap"] = "({}%)".format(self.number)
        args["time"] = "({}s)".format(self.number)

        if line[0] in ["*", "H"]:
            args["obj"] = "()"
            args["iinf"] = "()"
            args["depth"] = "()"

        if re.search(r"\*\s*\d+\+", line):
            args["obj"] = "()"
            args["ItCnt"] = "()"

        if re.search(r"Cuts: \d+", line):
            args["b_bound"] = r"(Cuts: \d+)"

        get = re.search(r"\*?\s*\d+\+?\s*\d+\s*(infeasible|cutoff|integral)", line)
        if get is not None:
            state = get.group(1)
            args["obj"] = "({})".format(state)
            if state in ["integral"]:
                pass
            else:
                args["iinf"] = "()"

        find = re.search(
            r"\s+{n}\s+{n_left}\s+{obj}\s+{depth}\s+{iinf}?\s+{b_int}?-?"
            r"\s+{b_bound}\s+{gap}?-?\s+{ItCnt}?-?\s+{time}".format(**args),
            line,
        )
        if not find:
            return None
        return find.groups()



# /usr/bin/python3
import re
import pandas as pd
import numpy as np
from .constants import (
    LpSolutionOptimal,
    LpSolutionIntegerFeasible,
    LpSolutionNoSolutionFound,
    solver_to_solution,
)


class LogFile(object):
    """
    This represents the log files that solvers return.
    We implement functions to get different information
    """

    name = None

    def __init__(self, path, **options):

        if options.get("content", False):
            content = path
        else:
            with open(path, "r") as f:
                content = f.read()

        self.path = path
        self.content = content
        self.number = r"-?[\de\.\+]+"
        self.numberSearch = r"({})".format(self.number)
        self.wordSearch = r"([\w, -]+)"

        self.solver_status_map = {}
        self.version_regex = ""
        self.progress_filter = ""
        self.progress_names = []
        self.options = options

    def apply_regex(
        self, regex, content_type=None, first=True, pos=None, num=None, **kwargs
    ):
        """
        regex is the regular expression to apply to the file contents.
        content_type are optional type casting for the results (int, float, str)
        if first is false, we take the raw output from the re.findall. Useful for progress table.
        num means, if there are multiple matches in re.findall, num tells which position to take out.
        if num=-1, we take the last one
        pos means the group we want to take out from all the groups of the relevant match.
        kwargs are additional parameters to the re.findall function
        :return: a list, a tuple or a single value with type "content_type"
        """
        solution = re.findall(regex, self.content, **kwargs)
        if solution is None:
            return None
        if not first:
            return solution
        if len(solution) == 0:
            return None
        if num is None:
            num = 0
        possible_tuple = solution[num]
        if type(possible_tuple) is str:
            # we force a tuple to deal with one string lists
            possible_tuple = (possible_tuple,)
            pos = 0
        func = {"float": float, "int": int, "str": str}
        if pos is not None:
            value = possible_tuple[pos]
            if content_type in func:
                return func[content_type](value)
            else:
                return possible_tuple[pos]
        if content_type is None:
            return possible_tuple
        if type(content_type) is not list:
            return [func[content_type](val) for val in possible_tuple]
        else:
            # each one has its own type.
            # by default, we use strings
            ct = ["str" for r in possible_tuple]
            for i, _c in enumerate(content_type):
                if _c in func:
                    ct[i] = _c
            return [func[ct[i]](val) for i, val in enumerate(possible_tuple)]

    def get_first_relax(self, progress) -> float | None:
        """
        scans the progress table for the initial relaxed solution
        :return: relaxation
        """
        bestBounds = progress.CutsBestBound[~progress.CutsBestBound.isna()]

        df_filter = bestBounds.apply(
            lambda x: re.search(r"^\s*{}$".format(self.number), x) is not None
        )
        if len(df_filter) > 0 and any(df_filter):
            return float(bestBounds[df_filter].iloc[0])
        return None

    def get_first_solution(self, progress):
        """
        scans the progress table for the initial integer solution
        :param progress: table with progress
        :return: dictionary with information on the moment of finding integer solution
        """
        vars_extract = ["Node", "NodesLeft", "BestInteger", "CutsBestBound"]
        df_filter = progress.BestInteger.fillna("").str.match(
            r"^\s*{}$".format(self.number)
        )
        # HACK: take out CBCs magic number (1e+50 for no integer solution found)
        df_filter_1e50 = progress.BestInteger.fillna("").str.match(r"^\s*1e\+50$")
        df_filter = np.all([df_filter, ~df_filter_1e50], axis=0)
        if len(df_filter) > 0 and any(df_filter):
            for col in vars_extract:
                floatSearch = r"[+-]?[\d]+(\.[\d]+)?([Ee][+-]?[\d]+)?"
                regex = "^({}).*$".format(floatSearch)
                progress[col] = progress[col].str.extract(regex)[[0]]
                # progress[col] = progress[col].str.replace('(?!{})'.format(self.number), '')
            return pd.to_numeric(progress[vars_extract][df_filter].iloc[0]).to_dict()
        return None

    @staticmethod
    def get_results_after_cuts(progress):
        """
        gets relaxed and integer solutions after the cuts phase has ended.
        :return: tuple of length two
        """
        df_filter = np.all(
            (
                progress.Node.str.match(r"^\*?H?\s*0"),
                progress.NodesLeft.str.match(r"^\+?H?\s*[012]"),
            ),
            axis=0,
        )

        # in case we have some progress after the cuts, we get those values
        # if not, we return None to later fill with best_solution and best_bound
        if not np.any(df_filter):
            return None, None
        sol_value = progress.BestInteger[df_filter].iloc[-1]
        relax_value = progress.CutsBestBound[df_filter].iloc[-1]

        # finally, we return the found values
        if sol_value and re.search(r"^\s*-?\d", sol_value):
            sol_value = float(sol_value)
        if relax_value and re.search(r"^\s*-?\d", relax_value):
            relax_value = float(relax_value)
        else:
            relax_value = None

        return relax_value, sol_value

    def get_log_info(self) -> dict:
        """
        Main function that builds the general output for every solver
        :return: a dictionary
        """
        version = self.get_version()
        matrix = self.get_matrix_dict()
        matrix_post = self.get_matrix_dict(post=True)
        status, objective, bound, gap_rel = self.get_stats()
        solver_status, solution_status = self.get_status_codes(status, objective)
        if bound is None and solution_status == LpSolutionOptimal:
            bound = objective
        if solution_status == LpSolutionOptimal:
            gap_rel = 0
        presolve = self.get_lp_presolve()
        time_out = self.get_time()
        nodes = self.get_nodes()
        root_time = self.get_root_time()
        if self.options.get("get_progress", True):
            progress = self.get_progress()
        else:
            progress = pd.DataFrame()
        first_relax = first_solution = None
        cut_info = self.get_cuts_dict(progress, bound, objective)

        if len(progress):
            first_relax = self.get_first_relax(progress)
            if solution_status in [LpSolutionIntegerFeasible, LpSolutionOptimal]:
                first_solution = self.get_first_solution(progress)

        return {
            "version": version,
            "solver": self.name,
            "status": status,
            "best_bound": bound,
            "best_solution": objective,
            "gap": gap_rel,
            "time": time_out,
            "matrix_post": matrix_post,
            "matrix": matrix,
            "cut_info": cut_info,
            "rootTime": root_time,
            "presolve": presolve,
            "first_relaxed": first_relax,
            "progress": progress,
            "first_solution": first_solution,
            "status_code": solver_status,
            "sol_code": solution_status,
            "nodes": nodes,
        }

    def get_cuts_dict(self, progress, best_bound, best_solution) -> dict:
        """
        builds a dictionary with all information regarding to the applied cuts
        :return: a dictionary
        """
        if not len(progress):
            return None
        cuts = self.get_cuts()
        if not cuts:
            # if no cuts were found, no cuts statistics are produced
            return {}
        cutsTime = self.get_cuts_time()
        after_cuts, sol_after_cuts = self.get_results_after_cuts(progress)
        if after_cuts is None:
            after_cuts = best_bound

        return {
            "time": cutsTime,
            "cuts": cuts,
            "best_bound": after_cuts,
            "best_solution": sol_after_cuts,
        }

    def get_matrix_dict(self, post=False) -> dict:
        """
        wrapper to both matrix parsers (before and after preprocess)
        :return: a dictionary with three elements or None
        """
        if post:
            matrix = self.get_matrix_post()
        else:
            matrix = self.get_matrix()

        if matrix is None:
            return None

        order = ["constraints", "variables", "nonzeros"]
        return {k: matrix[p] for p, k in enumerate(order)}

    def get_version(self) -> str:
        """
        gets the solver's version
        """
        return self.apply_regex(self.version_regex)

    def get_matrix(self) -> dict | None:
        return None

    def get_matrix_post(self) -> dict | None:
        return None

    def get_stats(self):
        return None, None, None, None

    def get_status_codes(self, status, obj) -> tuple[int, int]:
        """
        converts the status string into a solver code and a solution code
        to standardize the output among solvers
        :return: tuple of length 2
        """

        solver_status = self.solver_status_map.get(status)
        solution_status = solver_to_solution.get(solver_status)

        if obj is not None and solution_status == LpSolutionNoSolutionFound:
            solution_status = LpSolutionIntegerFeasible

        return solver_status, solution_status

    def get_cuts(self):
        return None

    def get_cuts_time(self) -> float | None:
        return None

    def get_lp_presolve(self) -> float | None:
        return None

    def get_time(self) -> float | None:
        return None

    def get_nodes(self) -> int | None:
        return None

    def get_root_time(self) -> float:
        return None

    def process_line(self, line):
        return None

    def get_progress(self) -> pd.DataFrame:
        """
        :return: pandas dataframe with 8 columns
        """
        lines = self.apply_regex(self.progress_filter, first=False, flags=re.MULTILINE)
        processed = [self.process_line(line) for line in lines]
        processed_clean = [p for p in processed if p is not None]
        progress = pd.DataFrame(processed_clean)
        if len(progress):
            progress.columns = self.progress_names
        return progress


if __name__ == "__main__":
    pass



import unittest
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import orloge as ol
import orloge.constants as c

DATADIR = os.path.join(os.path.dirname(__file__), "data")
ALMOST_KEYS = ["best_solution", "best_bound"]


class SolverTest(unittest.TestCase):

    fileinfo = {
        "93_01.txt": {
            "solver": "CPSAT",
            "version": "v9.3.10497",
            "time": 3.5908,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 15,
            "best_bound": 15,
            "status": "OPTIMAL",
        },
        "910_01.txt": {
            "solver": "CPSAT",
            "version": "v9.10.4067",
            "time": 30.4515,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionIntegerFeasible,
            "best_solution": 1450,
            "best_bound": 1515,
            "status": "FEASIBLE",
        },
        "cbc_0_objective": {
            "solver": "CBC",
            "version": "2.9.0",
            "time": 0.1,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 0,
            "best_bound": 0,
        },
        "cbc298-app1-2": {
            "solver": "CBC",
            "version": "2.9.8",
            "time": 7132.49,
            "nodes": 31867,
            "status_code": c.LpStatusTimeLimit,
            "sol_code": c.LpSolutionNoSolutionFound,
            "best_solution": None,
            "best_bound": -96.111,
        },
        "cbc298-ash608gpia-3col": {
            "solver": "CBC",
            "time": 224.71,
            "nodes": 0,
            "status_code": c.LpStatusInfeasible,
            "sol_code": c.LpSolutionInfeasible,
            "matrix": {"constraints": 24748, "variables": 3651, "nonzeros": 74244},
            "matrix_post": {"constraints": 18352, "variables": 3651, "nonzeros": 58254},
            "best_solution": None,
            "best_bound": None,
        },
        "cbc298-bab5": {
            "solver": "CBC",
            "time": 7194.79,
            "nodes": 162253,
            "status_code": c.LpStatusTimeLimit,
            "sol_code": c.LpSolutionIntegerFeasible,
            "best_solution": -104286.921,
            "best_bound": -111273.306,
        },
        "cbc298-dfn-gwin-UUM": {
            "solver": "CBC",
            "time": 725.37,
            "nodes": 378472,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 38752,
            "best_bound": 38752,
        },
        "cbc298-enlight14": {
            "solver": "CBC",
            "time": 7131.16,
            "nodes": 762406,
            "status_code": c.LpStatusTimeLimit,
            "sol_code": c.LpSolutionNoSolutionFound,
            "best_solution": None,
            "best_bound": 36.768,
        },
        "cbc298-satellites1-25": {
            "solver": "CBC",
            "time": 4511.76,
            "nodes": 51033,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -5,
            "best_bound": -5,
        },
        "cplex1271-app1-2": {
            "solver": "CPLEX",
            "version": "12.7.1.0",
            "nodes": 1415,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -41,
            "best_bound": -41,
        },
        "cplex1271-bab5": {
            "solver": "CPLEX",
            "nodes": 101234,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -106411.84,
            "best_bound": -106411.84,
        },
        "cplex1271-dfn-gwin-UUM": {
            "solver": "CPLEX",
            "nodes": 16132,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 38752,
            "best_bound": 38752,
        },
        "cplex1271-enlight14": {
            "solver": "CPLEX",
            "nodes": 0,
            "status_code": c.LpStatusInfeasible,
            "best_solution": None,
            "best_bound": None,
        },
        "cplex1271-satellites1-25": {
            "solver": "CPLEX",
            "nodes": 2942,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -5,
            "best_bound": -5,
        },
        "cplex1280-bab5": {
            "solver": "CPLEX",
            "time": 1551.53,
            "nodes": 51737,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -1.0641184010e05,
            "best_bound": -1.0641184010e05,
        },
        "cplex1280-enlight13": {
            "solver": "CPLEX",
            "time": 0.00,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 7.1000000000e01,
            "best_bound": 7.1000000000e01,
        },
        "cplex1280-enlight14": {
            "solver": "CPLEX",
            "time": 0.00,
            "nodes": 0,
            "status_code": c.LpStatusInfeasible,
            "best_solution": None,
            "best_bound": None,
        },
        "cplex1280-mine-90-10": {
            "solver": "CPLEX",
            "time": 924.71,
            "nodes": 388227,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -7.8430233763e08,
            "best_bound": -7.8430233763e08,
        },
        "cplex1280-rmine6": {
            "solver": "CPLEX",
            "time": 1437.13,
            "nodes": 540816,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -4.5718614000e02,
            "best_bound": -4.5718614070e02,
        },
        "cplex1280-satellites1-25": {
            "solver": "CPLEX",
            "time": 216.80,
            "nodes": 3896,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "matrix": {"constraints": 4658, "variables": 7759, "nonzeros": 50343},
            "matrix_post": {"constraints": 4214, "variables": 7088, "nonzeros": 45547},
            "cut_info": {
                "cuts": {
                    "GUB cover": 2,
                    "Clique": 7,
                    "Cover": 5,
                    "Implied bound": 21,
                    "Flow": 5,
                    "Mixed integer rounding": 19,
                    "Zero-half": 18,
                    "Gomory fractional": 2,
                },
                "time": 4.42,
                "best_bound": -20,
                "best_solution": 45,
            },
            "best_solution": -5.0000000000e00,
            "best_bound": -5.0000000000e00,
        },
        "cplex1280-tanglegram2": {
            "solver": "CPLEX",
            "time": 0.98,
            "nodes": 3,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 4.4300000000e02,
            "best_bound": 4.4300000000e02,
        },
        "cplex1280-fmpFP1": {
            "solver": "CPLEX",
            "time": 3605.16,
            "nodes": 28959,
            "status_code": c.LpStatusTimeLimit,
            "sol_code": c.LpSolutionIntegerFeasible,
            "matrix": {"constraints": 26699, "variables": 17492, "nonzeros": 248822},
            "matrix_post": {
                "constraints": 24199,
                "variables": 16436,
                "nonzeros": 200919,
            },
            "cut_info": {
                "cuts": {
                    "GUB cover": 14,
                    "Clique": 22,
                    "Cover": 13,
                    "Implied bound": 421,
                    "Flow": 86,
                    "Mixed integer rounding": 286,
                    "Zero-half": 199,
                    "Lift and project": 17,
                    "Gomory fractional": 49,
                },
                "time": 64.88,
                "best_bound": 75033.0540,
                "best_solution": None,
            },
            "first_solution": {
                "Node": 4101,
                "NodesLeft": 2714,
                "BestInteger": 119124.1217,
                "CutsBestBound": 75034.9046,
            },
            "best_solution": 7.8045498368e04,
            "best_bound": 7.5036256250e04,
        },
        "cplex1280-fmpFP3": {
            "solver": "CPLEX",
            "time": 0.17,
            "nodes": 0,
            "status_code": c.LpStatusInfeasible,
            "sol_code": c.LpSolutionInfeasible,
            "matrix": None,
            "matrix_post": None,
            "cut_info": None,
        },
        "cplex1280-fmpFP4": {
            "solver": "CPLEX",
            "time": 2.17,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "matrix": {"constraints": 6744, "variables": 4583, "nonzeros": 62949},
            "matrix_post": {"constraints": 6164, "variables": 4314, "nonzeros": 51540},
            "cut_info": {
                "cuts": {
                    "Clique": 1,
                    "Implied bound": 3,
                    "Flow": 0,
                    "Mixed integer rounding": 2,
                    "Zero-half": 3,
                    "Lift and project": 2,
                    "Gomory fractional": 6,
                },
                "time": None,
                "best_bound": 1.8003715497e04,
                "best_solution": 1.8041793e04,
            },
            "best_bound": 1.8003715497e04,
            "best_solution": 1.8041793e04,
        },
        "cplex1280-fmp_double_log": {
            "solver": "CPLEX",
            "time": 4.26,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "matrix": {"constraints": 3884, "variables": 8789, "nonzeros": 97473},
            "matrix_post": {"constraints": 3857, "variables": 8752, "nonzeros": 97138},
            "cut_info": {
                "cuts": {
                    "GUB cover": 6,
                    "Clique": 11,
                    "Cover": 1,
                    "Implied bound": 35,
                    "Flow": 6,
                    "Mixed integer rounding": 30,
                    "Zero-half": 18,
                    "Lift and project": 1,
                    "Gomory fractional": 27,
                },
                "time": None,
                "best_bound": 4.821578942e05,
                "best_solution": 4.821609305e05,
            },
            "first_solution": {
                "Node": 0,
                "NodesLeft": 0,
                "BestInteger": 482160.9305,
                "CutsBestBound": 482157.8942,
            },
            "best_solution": 4.821609305e05,
            "best_bound": 4.821578942e05,
        },
        "cplex1280-fmp-memoryerror": {
            "solver": "CPLEX",
            "version": "V12.8.0.0",
            "time": 2011.81,
            "nodes": None,
            "status_code": c.LpStatusMemoryLimit,
            "sol_code": c.LpSolutionIntegerFeasible,
            "matrix": {"constraints": 9203, "variables": 11818, "nonzeros": 492539},
            "matrix_post": {
                "constraints": 9194,
                "variables": 11803,
                "nonzeros": 492011,
            },
            "cut_info": {
                "cuts": {
                    "GUB cover": 26,
                    "Clique": 13,
                    "Cover": 473,
                    "Implied bound": 41,
                    "Flow": 14,
                    "Mixed integer rounding": 60,
                    "Zero-half": 39,
                    "Lift and project": 68,
                    "Gomory fractional": 41,
                },
                "time": 11.03,
                "best_bound": 704149.2079,
                "best_solution": 3798961.8736,
            },
            "first_solution": {
                "Node": 0,
                "NodesLeft": 0,
                "BestInteger": 4170639.0789,
                "CutsBestBound": 704149.1604,
            },
            "best_solution": 740618.6805,
            "best_bound": None,
        },
        "cplex1280-fmp-cutoff": {
            "solver": "CPLEX",
            "version": "V12.8.0.0",
            "time": 327.76,
            "nodes": 0,
            "status_code": c.LpStatusInfeasible,
            "sol_code": c.LpSolutionInfeasible,
            "matrix": {"constraints": 48938, "variables": 28780, "nonzeros": 3578287},
            "matrix_post": {
                "constraints": 48495,
                "variables": 28697,
                "nonzeros": 3555015,
            },
            "cut_info": {
                "cuts": {
                    "GUB cover": 19,
                    "Clique": 35,
                    "Cover": 1,
                    "Implied bound": 128,
                    "Flow": 23,
                    "Mixed integer rounding": 87,
                    "Zero-half": 28,
                    "Gomory fractional": 54,
                },
                "time": 327.59,
                "best_solution": None,
                "best_bound": None,
            },
            "best_solution": None,
            "best_bound": None,
        },
        "trivial_cplex_300": {
            "solver": "CPLEX",
            "version": "20.1.0.0",
            "time": 300.03,
            "matrix": {"constraints": 613, "variables": 428, "nonzeros": 23460},
            "matrix_post": {"constraints": 613, "variables": 428, "nonzeros": 23535},
        },
        "gurobi912-issue7": {
            "solver": "GUROBI",
            "time": 0.12,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "matrix": {"constraints": 3939, "variables": 2656, "nonzeros": 9894},
            "matrix_post": {"constraints": 2198, "variables": 1401, "nonzeros": 6851},
            "best_solution": 858381,
            "best_bound": 858381,
        },
        "gurobi700-app1-2": {
            "solver": "GUROBI",
            "time": 46.67,
            "nodes": 526,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "matrix": {"constraints": 53467, "variables": 26871, "nonzeros": 199175},
            "matrix_post": {
                "constraints": 35616,
                "variables": 22010,
                "nonzeros": 149085,
            },
            "best_solution": -41,
            "best_bound": -41,
        },
        "gurobi700-bab5": {
            "solver": "GUROBI",
            "time": 65.35,
            "nodes": 534,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -106411.84,
            "best_bound": -106411.84,
        },
        "gurobi700-dfn-gwin-UUM": {
            "solver": "GUROBI",
            "time": 388.98,
            "nodes": 170061,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 38752,
            "best_bound": 38752,
        },
        "gurobi700-enlight14": {
            "solver": "GUROBI",
            "time": 0.00,
            "nodes": 0,
            "status_code": c.LpStatusInfeasible,
            "best_solution": None,
            "best_bound": None,
        },
        "gurobi700-satellites1-25": {
            "solver": "GUROBI",
            "time": 59.70,
            "nodes": 1170,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -5,
            "best_bound": -5,
        },
        "gurobi800-bab5": {
            "solver": "GUROBI",
            "time": 25.52,
            "nodes": 1,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": -1.064118401000e05,
            "best_bound": -1.064118401000e05,
        },
        "gurobi800-enlight13": {
            "solver": "GUROBI",
            "time": 0.00,
            "nodes": 0,
            "status_code": c.LpStatusSolved,
            "sol_code": c.LpSolutionOptimal,
            "best_solution": 7.100000000000e01,
            "best_bound": 7.100000000000e01,
        },
        "gurobi800-enlight14": {
            "solver": "GUROBI",
            "time": 0.00,
            "nodes": 0,
            "status_code": c.LpStatusInfeasible,
            "best_solution": None,
            "best_bound": None,
        },
        "gurobi800-mine-90-10": {
            "solver": "GUROBI",
            "time": 769.45,
            "nodes": 1181932,
            "status_code": c.LpStatusSolved,
            "best_solution": -7.843023376332e08,
            "best_bound": -7.843023376332e08,
        },
        "gurobi800-satellites1-25": {
            "solver": "GUROBI",
            "time": 40.86,
            "nodes": 380,
            "status_code": c.LpStatusSolved,
            "best_solution": -5.000000000000e00,
            "best_bound": -5.000000000000e00,
        },
        "gurobi800-tanglegram2": {
            "solver": "GUROBI",
            "time": 0.57,
            "nodes": 1,
            "status_code": c.LpStatusSolved,
            "matrix": {"constraints": 8980, "variables": 4714, "nonzeros": 26940},
            "matrix_post": {"constraints": 7976, "variables": 4058, "nonzeros": 23928},
            "cut_info": {
                "cuts": {"Zero half": 5},
                "time": 0,
                "best_bound": 443,
                "best_solution": 443,
            },
            "best_solution": 4.430000000000e02,
            "best_bound": 4.430000000000e02,
        },
        "gurobi901-noCuts": {
            "solver": "GUROBI",
            "time": 68.88,
            "nodes": 1,
            "status_code": c.LpStatusSolved,
            "matrix": {"constraints": 537927, "variables": 381096, "nonzeros": 1946442},
            "matrix_post": {
                "constraints": 74769,
                "variables": 10145,
                "nonzeros": 314632,
            },
            "cut_info": {},
            "best_solution": 36,
            "best_bound": 36,
        },
        # "xpress300103-app1-2": {
        #     'solver': "XPRESS",
        #     'time': 29,
        #     'nodes': 265,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':-41,
        #     'best_bound':-41},
        # "xpress300103-bab5": {
        #     'solver': "XPRESS",
        #     'time': 7200,
        #     'nodes': 71853,
        #     'status_code': c.LpStatusTimeLimit,
        #     'best_solution':-106411.84,
        #     'best_bound':-106701.8161},
        # "xpress300103-dfn-gwin-UUM": {
        #     'solver': "XPRESS",
        #     'time':181,
        #     'nodes': 158849,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': 38752,
        #     'best_bound': 38748},
        # "xpress300103-enlight14": {
        #     'solver': "XPRESS",
        #     'time': 0,
        #     'nodes': 0,
        #     'status_code': c.LpStatusInfeasible,
        #     'best_solution': 1e+40,
        #     'best_bound': 1e+40},
        # "xpress300103-satellites1-25": {
        #     'solver': "XPRESS",
        #     'time': 228,
        #     'nodes': 5937,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':-5,
        #     'best_bound':-5},
        # "xpress330103-bab5": {
        #     'solver': "XPRESS",
        #     'time': 7200,
        #     'nodes': 24575,
        #     'status_code': c.LpStatusTimeLimit,
        #     'best_solution': -106361.339104,
        #     'best_bound': -106600.8564439218171173707},
        # "xpress330103-enlight13": {
        #     'solver': "XPRESS",
        #     'time': 0,
        #     'nodes': 0,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': 71,
        #     'best_bound': 71},
        # "xpress330103-enlight14": {
        #     'solver': "XPRESS",
        #     'time': 84,
        #     'nodes': 1,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': 52200,
        #     'best_bound': 52200.00000000001455191523},
        # "xpress330103-mine-90-10": {
        #     'solver': "XPRESS",
        #     'time': 5,
        #     'nodes': 505,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': -566395707.871,
        #     'best_bound': -566395707.8708435297012329},
        # "xpress330103-satellites1-25": {
        #     'solver': "XPRESS",
        #     'time': 99,
        #     'nodes': 7917,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': -5.00000000001,
        #     'best_bound': -5.000000000005489830812166},
        # "xpress330103-tanglegram2": {
        #     'solver': "XPRESS",
        #     'time': 0,
        #     'nodes': 5,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': 443,
        #     'best_bound': 443},
        # "mipcl131-app1-2": {
        #     'solver': "MIPCL",
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':-41,
        #     'best_bound':-41},
        # "mipcl131-bab5": {
        #     'solver': "MIPCL",
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':-106411.84,
        #     'best_bound':-106411.84},
        # "mipcl131-dfn-gwin-UUM": {
        #     'solver': "MIPCL",
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': 38752,
        #     'best_bound': 38752},
        # "mipcl131-enlight14": {
        #     'solver': "MIPCL",
        #     'status_code': c.LpStatusInfeasible,
        #     'best_solution': None,
        #     'best_bound': None},
        # "mipcl131-satellites1-25": {
        #     'solver': "MIPCL",
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':-5,
        #     'best_bound':-5},
        # "mipcl152-bab5": {
        #     'solver': "MIPCL",
        #     'time': 514.635,
        #     'nodes': 264,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': -106411.8401,
        #     'best_bound': -106411.8401},
        # "mipcl152-bnatt350": {
        #     'solver': "MIPCL",
        #     'time': 7199.330,
        #     'nodes': 117400,
        #     'status_code': c.LpStatusTimeLimit,
        #     'best_solution': None,
        #     'best_bound': -0.0000},
        # "mipcl152-enlight13": {
        #     'solver': "MIPCL",
        #     'time': 70.477,
        #     'nodes': 280,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': 71.0000,
        #     'best_bound': 71.0000},
        # "mipcl152-enlight14": {
        #     'solver': "MIPCL",
        #     'time': 280.405,
        #     'nodes': 1025,
        #     'status_code': c.LpStatusInfeasible,
        #     'best_solution': None,
        #     'best_bound': None},
        # "mipcl152-mine-90-10": {
        #     'solver': "MIPCL",
        #     'time': 7.370,
        #     'nodes': 1930,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': -566395707.8708,
        #     'best_bound': -566395707.8708},
        # "mipcl152-satellites1-25": {
        #     'solver': "MIPCL",
        #     'time': 2999.199,
        #     'nodes': 14031,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': -5.0000,
        #     'best_bound': -5.0000},
        # "mipcl152-tanglegram2": {
        #     'solver': "MIPCL",
        #     'time': 35.829,
        #     'nodes': 1,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution': 443.0000,
        #     'best_bound': 443.0000},
        # "scip400-infeasible": {
        #     'solver': "SCIP",
        #     'nodes': 1,
        #     'time': 0.02,
        #     Key.GitHash: "ea0b6dd",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "3.0.0",
        #     "SpxGitHash": "b0cccbd",
        #     'status_code': c.LpStatusInfeasible,
        #     'best_solution':+1.00000000000000e+20,
        #     'best_bound':+1.00000000000000e+20},
        # "scip400-memorylimit": {
        #     'solver': "SCIP",
        #     'nodes': 1778883,
        #     'time': 6807.85,
        #     Key.GitHash: "ea0b6dd",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "3.0.0",
        #     "SpxGitHash": "b0cccbd",
        #     'status_code': c.LpStatusMemoryLimit,
        #     'best_solution':+1.49321500000000e+06,
        #     'best_bound':+1.49059347656250e+06},
        # "scip400-optimal": {
        #     'solver': "SCIP",
        #     'nodes': 126,
        #     'time': 0.79,
        #     Key.GitHash: "dd19a7b",
        #     "mode": "optimized",
        #     "LPSolver": "CPLEX",
        #     "LPSolverVersion": "12.6.0.0",
        #     "SpxGitHash": None,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':+3.36000000000000e+03,
        #     'best_bound':+3.36000000000000e+03},
        # "scip400-timelimit": {
        #     'solver': "SCIP",
        #     'nodes': 101678,
        #     'time': 600.00,
        #     Key.GitHash: "dd19a7b",
        #     "mode": "optimized",
        #     "LPSolver": "CPLEX",
        #     "LPSolverVersion": "12.6.0.0",
        #     "SpxGitHash": None,
        #     'status_code': c.LpStatusTimeLimit,
        #     'best_solution':+1.16800000000000e+03,
        #     'best_bound':+1.13970859166290e+03},
        # "scip400-crashed": {
        #     'solver': "SCIP",
        #     'nodes': None,
        #     'time': None,
        #     Key.GitHash: "dd19a7b",
        #     "mode": "optimized",
        #     "LPSolver": "CPLEX",
        #     "LPSolverVersion": "12.6.0.0",
        #     "SpxGitHash": None,
        #     'status_code': Key.SolverStatusCodes.Crashed,
        #     'best_solution': None,
        #     'best_bound': None},
        # "scip500-bab5": {
        #     'solver': "SCIP",
        #     'time': 7200.00,
        #     'nodes': 120255,
        #     Key.GitHash: "3bbd232",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "3.1.0",
        #     "SpxGitHash": "5147d37",
        #     'status_code': c.LpStatusTimeLimit,
        #     'best_solution':-1.06411840100000e+05,
        #     'best_bound':-1.06736016607273e+05},
        # "scip500-enlight13": {
        #     'solver': "SCIP",
        #     'time': 0.01,
        #     'nodes': 1,
        #     Key.GitHash: "3bbd232",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "3.1.0",
        #     "SpxGitHash": "5147d37",
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':+7.10000000000000e+01,
        #     'best_bound':+7.10000000000000e+01},
        # "scip500-enlight14": {
        #     'solver': "SCIP",
        #     'time': 0.01,
        #     'nodes': 1,
        #     Key.GitHash: "3bbd232",
        #     "mode": "optimized",
        #     "LPSolver": "CPLEX",
        #     "LPSolverVersion": "12.8.0.0",
        #     "SpxGitHash": None,
        #     'status_code': c.LpStatusInfeasible,
        #     'best_solution':+1.00000000000000e+20,
        #     'best_bound':+1.00000000000000e+20},
        # "scip500-mine-90-10": {
        #     'solver': "SCIP",
        #     'time': 45.86,
        #     'nodes': 1483,
        #     Key.GitHash: "3bbd232",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "3.1.0",
        #     "SpxGitHash": "5147d37",
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':-5.66395707870830e+08,
        #     'best_bound':-5.66395707870830e+08},
        # "scip500-satellites1-25": {
        #     'solver': "SCIP",
        #     'time': 454.56,
        #     'nodes': 424,
        #     Key.GitHash: "3bbd232",
        #     "mode": "optimized",
        #     "LPSolver": "CPLEX",
        #     "LPSolverVersion": "12.8.0.0",
        #     "SpxGitHash": None,
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':-4.99999999999994e+00,
        #     'best_bound':-4.99999999999994e+00},
        # "scip500-tanglegram2": {
        #     'solver': "SCIP",
        #     'time': 5.82,
        #     'nodes': 3,
        #     Key.GitHash: "3bbd232",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "3.1.0",
        #     "SpxGitHash": "5147d37",
        #     'status_code': c.LpStatusSolved,
        #     'best_solution':+4.43000000000000e+02,
        #     'best_bound':+4.43000000000000e+02},
        # "scip501-in-memlim": {
        #     'solver': "SCIP",
        #     'time': 63.27,
        #     'nodes': 0,
        #     Key.GitHash: "b1c222a",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "3.1.1",
        #     "SpxGitHash": "ab921a5",
        #     'status_code': c.LpStatusMemoryLimit,
        #     'best_solution': 1e+20,
        #     'best_bound': None},
        # "scip600-in-memlim": {
        #     'solver': "SCIP",
        #     'time': 70.67,
        #     'nodes': 1,
        #     Key.GitHash: "ce154f0",
        #     "mode": "optimized",
        #     "LPSolver": "SoPlex",
        #     "LPSolverVersion": "4.0.0",
        #     "SpxGitHash": "e0d3842",
        #     'status_code': c.LpStatusMemoryLimit,
        #     'best_solution': 1e+20,
        #     'best_bound': 13.0},
    }

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testData(self):
        for filename, contents in self.fileinfo.items():
            file = self.getFileName(filename)
            solver = contents["solver"]
            data = ol.get_info_solver(file, solver)
            for key, value in contents.items():
                if key not in data:
                    print("not checking: {} in {}".format(key, filename))
                    continue
                msg = "{0} has {1} as {2}, should be {3}".format(
                    filename, data[key], key, value
                )
                if key in ALMOST_KEYS:
                    self.assertAlmostEqual(data[key], value, delta=1e-1, msg=msg)
                else:
                    self.assertEqual(data[key], value, msg=msg)

    @staticmethod
    def getFileName(basename):
        basename, ext = os.path.splitext(basename)
        if ext == "":
            ext = ".out"
        return os.path.join(DATADIR, f"{basename}{ext}")


if __name__ == "__main__":
    unittest.main()



from .base import LogFile
from .constants import (
    LpStatusMemoryLimit,
    LpStatusSolved,
    LpStatusInfeasible,
    LpStatusTimeLimit,
    LpStatusUnbounded,
    LpStatusNotSolved,
)
import re


class CBC(LogFile):
    def __init__(self, path, **options):
        super().__init__(path, **options)
        self.name = "CBC"
        self.solver_status_map = {
            "Optimal solution found": LpStatusSolved,
            "Problem is infeasible": LpStatusInfeasible,
            "Stopped on time limit": LpStatusTimeLimit,
            "Problem proven infeasible": LpStatusInfeasible,
            "Problem is unbounded": LpStatusUnbounded,
            "Pre-processing says infeasible or unbounded": LpStatusInfeasible,
            "** Current model not valid": LpStatusNotSolved,
        }
        self.version_regex = r"Version: (\S+)"
        self.progress_names = [
            "Node",
            "NodesLeft",
            "BestInteger",
            "CutsBestBound",
            "Time",
        ]
        self.progress_filter = r"(^Cbc0010I.*$)"

    def get_cuts(self):
        # TODO
        pass

    def get_matrix(self):
        regex = r"Problem .+ has {0} rows, {0} columns and {0} elements".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="int")

    def get_matrix_post(self):
        regex = (
            r"Cgl0004I processed model has {0} rows, {0} columns \(\d+ integer "
            r"\(\d+ of which binary\)\) and {0} elements".format(self.numberSearch)
        )
        return self.apply_regex(regex, content_type="int")

    def get_stats(self):

        regex = "Result - {}".format(self.wordSearch)
        status = self.apply_regex(regex, pos=0)
        if status is None:
            # no solution found, I still want the status
            for k in self.solver_status_map.keys():
                if self.apply_regex(re.escape(k)):
                    return k, None, None, None
        else:
            status = status.strip()
        regex = r"best objective {0}( \(best possible {0}\))?, took {1} iterations and {1} nodes \({1} seconds\)".format(
            self.numberSearch, self.number
        )
        solution = self.apply_regex(regex)

        if solution is None:
            return None, None, None, None

        # if solution[0] == '1e+050':
        if self.apply_regex("No feasible solution found"):
            objective = None
        else:
            objective = float(solution[0])

        if solution[2] is None or solution[2] == "":
            bound = objective
        else:
            bound = float(solution[2])

        gap_rel = None
        if objective is not None and objective != 0:
            gap_rel = abs(objective - bound) / abs(objective) * 100

        return status, objective, bound, gap_rel

    def get_cuts_time(self):
        # TODO
        return None

    def get_lp_presolve(self):
        # TODO
        return None

    def get_time(self):
        regex = r"Total time \(CPU seconds\):\s*{}".format(self.numberSearch)
        stats = self.apply_regex(regex, content_type="float", pos=0)
        return stats

    def get_nodes(self):
        regex = r"Enumerated nodes:\s*{}".format(self.numberSearch)
        return self.apply_regex(regex, content_type="int", pos=0)

    def get_root_time(self):
        # TODO
        return None

    def process_line(self, line):

        keys = ["n", "n_left", "b_int", "b_bound", "time"]
        args = {k: self.numberSearch for k in keys}
        find = re.search(
            r"Cbc0010I After {n} nodes, {n_left} on tree, {b_int} best solution, "
            r"best possible {b_bound} \({time} seconds\)".format(**args),
            line,
        )
        if not find:
            return None
        return find.groups()



__all__ = ["base", "cplex", "gurobi", "cbc", "cpsat"]

from .cplex import CPLEX
from .gurobi import GUROBI
from .cbc import CBC
from .cpsat import CPSAT

__map = dict(CPLEX=CPLEX, GUROBI=GUROBI, CBC=CBC, CPSAT=CPSAT)


def get_info_solver(path, solver, **options):
    my_solver = get_solver(solver)
    if my_solver is None:
        raise ValueError(f"solver {solver} is not recognized")
    log = my_solver(path, **options)
    return log.get_log_info()


def get_solver(solver):
    my_solver = __map.get(solver)
    if my_solver is None:
        raise ValueError(f"solver {solver} is not recognized")
    return my_solver



from .base import LogFile
from .constants import (
    LpStatusMemoryLimit,
    LpStatusSolved,
    LpStatusInfeasible,
    LpStatusTimeLimit,
    LpStatusUnbounded,
    LpStatusNotSolved,
)
import re
import numpy as np


class CPLEX(LogFile):
    # Reference:
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.6.3/ilog.odms.cplex.help/CPLEX/UsrMan/topics/discr_optim/mip/para/52_node_log.html
    name = "CPLEX"

    def __init__(self, path, **options):
        super().__init__(path, **options)
        self.solver_status_map = {
            "MIP - Memory limit exceeded": LpStatusMemoryLimit,
            "MIP - Integer optimal": LpStatusSolved,
            "MIP - Integer infeasible.": LpStatusInfeasible,
            "MIP - Time limit exceeded": LpStatusTimeLimit,
            "MIP - Integer unbounded": LpStatusUnbounded,
            "MIP - Integer infeasible or unbounded": LpStatusInfeasible,
            "CPLEX Error  1001: Out of memory": LpStatusMemoryLimit,
            "No file read": LpStatusNotSolved,
        }

        self.version_regex = [
            r"Welcome to IBM\(R\) ILOG\(R\) CPLEX\(R\) Interactive Optimizer (\S+)",
            r"Log started \((\S+)\)",
            r"Version identifier: (\S+)",
        ]
        self.header_log_start = ["Welcome to IBM", "Log started"]
        self.progress_names = [
            "Node",
            "NodesLeft",
            "Objective",
            "IInf",
            "BestInteger",
            "CutsBestBound",
            "ItpNode",
            "Gap",
        ]
        self.progress_filter = r"(^[\*H]?\s*\d.*$)"
        # in case of multiple logs in the same file,
        # we choose to get the last one.
        self.content = self.clean_before_last_log()

    def clean_before_last_log(self):
        options = self.header_log_start
        for opt in options:
            # this finds the last occurence of the string
            pos = self.content.rfind(opt)
            if pos != -1:
                return self.content[pos:]
        return self.content

    def get_version(self):
        result = None
        for reg in self.version_regex:
            result = self.apply_regex(reg)
            if result:
                return result
        return result

    def get_stats(self):
        status = self.get_status()
        objective = self.get_objective()
        bound, gap_abs, gap_rel = self.get_gap()
        return status, objective, bound, gap_rel

    def get_status(self):
        for k in self.solver_status_map.keys():
            search_string = re.escape(k)
            if self.apply_regex(search_string):
                return k

    def get_objective(self):
        """
        :return: tuple of length 2
        """
        regex = r"Objective\s+=\s+{0}\s*\n".format(self.numberSearch)
        result = self.apply_regex(regex, flags=re.MULTILINE)
        if result is not None:
            result = float(result)
        return result

    def get_gap(self):
        """
        :return: tuple of length 3: bound, absolute gap, relative gap
        """
        regex = r"Current MIP best bound =\s+{0} \(gap = {0}, {0}%\)".format(
            self.numberSearch
        )
        result = self.apply_regex(regex, content_type="float")
        if result is None:
            return None, None, None
        return result

    def get_matrix(self):
        """
        :return: tuple of length 3
        """
        # TODO: this is not correctly calculated: we need to sum the change to the initial to get
        #  the original.
        regex = r"Reduced MIP has {0} rows, {0} columns, and {0} nonzeros".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="int", num=0)

    def get_matrix_post(self):
        """
        :return: tuple of length 3
        """
        regex = r"Reduced MIP has {0} rows, {0} columns, and {0} nonzeros".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="int", num=-1)

    def get_cuts(self):
        """
        :return: dictionary of cuts
        """
        regex = r"{1} cuts applied:  {0}".format(self.numberSearch, self.wordSearch)
        result = self.apply_regex(regex, first=False)
        if result is None:
            return None
        return {k[0]: int(k[1]) for k in result}

    def get_lp_presolve(self):
        """
        :return: tuple  of length 3
        """
        # TODO: this is not correctly calculated:
        #  we need to sum the two? preprocessings in my cases
        regex = r"Presolve time = {0} sec. \({0} ticks\)".format(self.numberSearch)
        time = self.apply_regex(regex, pos=0, content_type="float")

        regex = r"LP Presolve eliminated {0} rows and {0} columns".format(
            self.numberSearch
        )
        result = self.apply_regex(regex, content_type="int")
        if result is None:
            result = None, None
        return {"time": time, "rows": result[0], "cols": result[1]}

    def get_time(self):
        regex = r"Solution time =\s+{0} sec\.\s+Iterations = {0}\s+Nodes = {0}".format(
            self.numberSearch
        )
        result = self.apply_regex(regex, content_type="float", pos=0)
        if result is not None:
            return result
        regex = r"Total \(root\+branch&cut\) =\s+{0} sec\. \({0} ticks\)".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="float", pos=0)

    def get_nodes(self):
        regex = r"Solution time =\s+{0} sec\.\s+Iterations = {0}\s+Nodes = {0}".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="float", pos=2)

    def get_root_time(self):
        regex = r"Root relaxation solution time = {0} sec\. \({0} ticks\)".format(
            self.numberSearch
        )
        return self.apply_regex(regex, pos=0, content_type="float")

    def get_cuts_time(self):
        regex = r"Elapsed time = {0} sec\. \({0} ticks, tree = {0} MB, solutions = {0}\)".format(
            self.numberSearch
        )
        return self.apply_regex(regex, content_type="float", pos=0)

    def process_line(self, line):
        keys = ["n", "n_left", "obj", "iinf", "b_int", "b_bound", "ItCnt", "gap"]
        args = {k: self.numberSearch for k in keys}
        args["gap"] = "({}%)".format(self.number)

        if re.search(r"\*\s*\d+\+", line):
            args["obj"] = "()"
            args["ItCnt"] = "()"
            args["iinf"] = "()"

        # TODO: maybe include explicit optioms: Impl Bds, Cuts, ZeroHalf, Flowcuts,
        if re.search(r"[a-zA-Z\s]+: \d+", line):
            args["b_bound"] = r"([a-zA-Z\s]+: \d+)"

        get = re.search(r"\*?\s*\d+\+?\s*\d+\s*(infeasible|cutoff|integral)", line)
        if get is not None:
            state = get.group(1)
            args["obj"] = "({})".format(state)
            if state in ["integral"]:
                args["iinf"] = "(0)"
            else:
                args["iinf"] = "()"
            if state in ["cutoff", "infeasible"]:
                args["b_bound"] += "?"

        find = re.search(
            r"\s*{n}\s*{n_left}\s+{obj}\s+{iinf}?\s+{b_int}?\s+{b_bound}\s+{ItCnt}\s*{gap}?".format(
                **args
            ),
            line,
        )
        if not find:
            return None
        return find.groups()

    def get_progress(self):
        progress = super().get_progress()
        if len(progress):
            try:
                times = self.get_time_column()
                progress["Time"] = times
            except TypeError:
                progress["Time"] = None
        return progress

    def get_time_column(self):
        """
        :return: Time column with same length as progress dataframe.
        """
        regex1 = self.progress_filter
        args = [self.numberSearch for l in range(4)]
        regex = (
            r"Elapsed time = {} sec. \({} ticks, tree = {} MB, solutions = {}\)".format(
                *args
            )
        )
        end_time = self.get_time()
        table_start_rx = r"\s*Node"
        table_start = False
        i = 0
        time = [(i, 0)]
        for l in self.content.split("\n"):
            if re.search(table_start_rx, l):
                table_start = True
            if not table_start:
                continue
            if re.search(regex1, l):
                i += 1
                continue
            result = re.search(regex, l)
            if not result:
                continue
            data = result.groups()
            time.append((i, float(data[0])))
        time.append((i, end_time))
        x, y = zip(*time)
        numbers = np.interp(xp=x, fp=y, x=range(1, x[-1] + 1)).round(2)
        # we coerce to string to match the other solvers output:
        return numbers.astype("str")



from cpsat_logutils.blocks import SearchProgressBlock, SolverBlock, ResponseBlock
from cpsat_logutils.blocks.search_progress import ModelEvent
from .base import LogFile
import cpsat_logutils as cpsatlog
import pandas as pd
from .constants import (
    LpStatusMemoryLimit,
    LpStatusSolved,
    LpStatusInfeasible,
    LpSolutionIntegerFeasible,
    LpStatusTimeLimit,
    LpStatusUnbounded,
    LpStatusNotSolved,
    LpSolutionOptimal,
    LpSolutionInfeasible,
    LpSolutionNoSolutionFound,
)


class CPSAT(LogFile):
    my_parser: cpsatlog.LogParser
    name = "CPSAT"

    def __init__(self, path, **options):
        super().__init__(path, **options)

        self.my_parser = cpsatlog.LogParser(self.content)

    def get_progress(self):
        """
        Builds a pandas DataFrame with the search progress.
        It assumes events are sorted by time.
        It fills with the previous result when no new information is available.
        The return DataFrame has the following columns:
        [the names of columns match the ones in the orloge library,
        this may not be very relevant to cp-sat]
        - Time: The time of the event.
        - CutsBestBound: The best bound found at the time.
        - BestInteger: The best integer found at the time.
        - Gap: The gap between the best bound and the best integer.
        - NumVars: The number of variables before the event.
        - RemVars: The number of remaining variables after the event.
        - NumCons: The number of constraints before the event.
        - RemCons: The number of remaining constraints after the event.
        """
        progress_block = self.my_parser.get_block_of_type_or_none(SearchProgressBlock)
        events = progress_block._parse_events()
        gap = obj = bound = None
        cons = new_cons = var = new_var = None
        my_table = []
        for e in events:
            if not isinstance(e, ModelEvent):
                bound = e.bound
                obj = e.obj
                gap = e.get_gap()
            else:
                var = e.vars
                new_var = e.vars_remaining
                cons = e.constr
                new_cons = e.constr_remaining
            my_table.append((e.time, bound, obj, gap, var, new_var, cons, new_cons))
        col_names = [
            "Time",
            "CutsBestBound",
            "BestInteger",
            "Gap",
            "NumVars",
            "RemVars",
            "NumCons",
            "RemCons",
        ]

        return pd.DataFrame.from_records(my_table, columns=col_names)

    def get_first_relax(self, progress):
        return None

    def get_nodes(self):
        return 0

    def get_time(self):
        my_block = self.my_parser.get_block_of_type_or_none(ResponseBlock)
        data = my_block.to_dict()
        return float(data["usertime"])

    def get_cuts(self):
        pass

    def get_version(self):
        my_block = self.my_parser.get_block_of_type_or_none(SolverBlock)
        return my_block.get_version()

    def get_cuts_dict(self, progress, bound, objective):
        return None

    def get_stats(self):
        # status, objective, bound, gap_rel
        my_block = self.my_parser.get_block_of_type_or_none(ResponseBlock)
        data = my_block.to_dict()
        gap = my_block.get_gap()
        return data["status"], float(data["objective"]), float(data["best_bound"]), gap

    def get_status_codes(self, status, obj):
        _map_status = dict(
            OPTIMAL=LpStatusSolved,
            FEASIBLE=LpStatusSolved,
            INFEASIBLE=LpStatusInfeasible,
            UNBOUNDED=LpStatusUnbounded,
            TIME_LIMIT=LpStatusTimeLimit,
            MEMORY_LIMIT=LpStatusMemoryLimit,
            UNKNOWN=LpStatusNotSolved,
        )
        _map_sol_status = dict(
            OPTIMAL=LpSolutionOptimal,
            FEASIBLE=LpSolutionIntegerFeasible,
            INFEASIBLE=LpSolutionInfeasible,
            UNKNOWN=LpSolutionNoSolutionFound,
        )
        return _map_status.get(status, LpStatusNotSolved), _map_sol_status.get(
            status, LpSolutionNoSolutionFound
        )

    def get_first_solution(self, progress):
        return None
